/*  Jcurzez - A free Java implementation of curses
 *  Copyright (C) 2001, 2006  Thomas G. Girard <thomas.g.girard@free.fr>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.nongnu.savannah.jcurzez;

import atomicsets.OrderedLock;

/**
 * <code>FillingPen</code> class has a {@link jcurzez.Cell cell} attached,
 * which represents the blank cell. It is used for filling new room created
 * when one {@link jcurzez.AbstractWindow#deleteChar deletes a character},
 * {@link jcurzez.AbstractWindow#insertLine inserts a new line}, and so
 * on. Anytime this filling pen is updated, the empty cell is updated.
 * <p>
 * <i>N.B: this class is package private.</i>
 *
 * @author Thomas Girard
 * @version 1.0.0
 *
 * @see jcurzez.AbstractWindow AbstractWindow
 * @see jcurzez.Cell Cell
 */
class FillingPen extends Pen {
    /**
     * The blank cell. Updated whenever <code>this</code> is changed.
     */
    final Cell blankCell;

    /**
	 * Creates a new <code>FillingPen</code>. It defaults to white on black, with normal attributes.
	 */
	FillingPen() {
		this(new OrderedLock());
	}

	/**
     * Creates a new <code>FillingPen</code>.
     * It defaults to white on black, with normal attributes.
     */
    FillingPen(OrderedLock P) {
	this((int) ' ', ForegroundColor.WHITE,
	     BackgroundColor.BLACK, Attribute.NORMAL, P);
    }

    /**
	 * Creates a new <code>FillingPen</code>, with specified pen.
	 * @param p  the pen to use for filling.
	 */
	FillingPen(Pen p) {
		this(p, new OrderedLock());
	}

	/**
     * Creates a new <code>FillingPen</code>, with specified pen.
     *
     * @param     p the pen to use for filling.
     */
    FillingPen(/*unitfor(P)*/ Pen p, OrderedLock P) {
		this((int) ' ', p.getForegroundColor_internal(), p.getBackgroundColor_internal(), p.getAttribute_internal(), P);
		synchronized (p.getLockForP()) {
		}
	}

    /**
	 * Creates a new <code>FillingPen</code> with specified attributes.
	 * @param c  the character to use for filling (should be <code>' ' </code>).
	 * @param foreground  the foreground color to use for filling.
	 * @param background  the background color to use for filling.
	 * @param attribute  the attribute to use for filling.
	 */
	FillingPen(int c, ForegroundColor foreground, BackgroundColor background,
			Attribute attribute) {
		this(c, foreground, background, attribute, new OrderedLock());
	}

	/**
     * Creates a new <code>FillingPen</code> with specified attributes.
     *
     * @param     c the character to use for filling
     *              (should be <code>' ' </code>).
     * @param     foreground the foreground color to use for filling.
     * @param     background the background color to use for filling.
     * @param     attribute the attribute to use for filling.
     */
    FillingPen(int c, ForegroundColor foreground,
		      BackgroundColor background,
		      Attribute attribute, OrderedLock P) {
	super(foreground, background, attribute, P);

	//hasChanged is always true because blankCell is used for new cells.
	this.blankCell = new Cell(c, attribute, background, foreground, true);
    }

    /**
     * Modifies {@link #blankCell}.
     */
    public void setBackgroundColor_internal(BackgroundColor background) {
	blankCell.background = background;
	super.setBackgroundColor(background);
    }

	/**
	 * Modifies  {@link #blankCell} .
	 */
	public void setBackgroundColor(BackgroundColor background) {
	blankCell.background = background;
	super.setBackgroundColor(background);
    }

    /**
     * Modifies {@link #blankCell}.
     */
    public void setForegroundColor_internal(ForegroundColor foreground) {
	blankCell.foreground = foreground;
	super.setForegroundColor(foreground);
    }

	/**
	 * Modifies  {@link #blankCell} .
	 */
	public void setForegroundColor(ForegroundColor foreground) {
	blankCell.foreground = foreground;
	super.setForegroundColor(foreground);
    }

    /**
     * Modifies {@link #blankCell}.
     */
    public void setAttribute_internal(Attribute attribute) {
	blankCell.attribute = attribute;
	super.setAttribute(attribute);
    }

	/**
	 * Modifies  {@link #blankCell} .
	 */
	public void setAttribute(Attribute attribute) {
	blankCell.attribute = attribute;
	super.setAttribute(attribute);
    }
}
