/*  Jcurzez - A free Java implementation of curses
 *  Copyright (C) 2001, 2006  Thomas G. Girard <thomas.g.girard@free.fr>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.nongnu.savannah.jcurzez;

/**
 * Every {@link AbstractWindow window} has a cursor. The cursor can be moved and
 * hidden.
 * 
 * @author Thomas Girard
 * @version 1.0.0
 * 
 * @see jcurzez.Pen
 */
public final /*internal*/ class Cursor {
	
	/*atomicset(C)*/
	
	/** The rectangle this cursor is attached to. */
	final Rectangle rectangle /*R=this.C*/;

	/** visibility of the cursor. */
	private int visibility = NORMAL;

	/** x coordinate of the cursor. */
	/*atomic(C)*/int x;

	/** y coordinate of the cursor. */
	/*atomic(C)*/int y;

	// not needed with aj
	//AbstractWindow aw;

	/**
	 * Constructs a new cursor for specified <code>rectangle</code>, with
	 * initial position <code>(initialX, initialY)</code>.
	 * 
	 * @param rectangle
	 *            the rectangle this cursor is attached to.
	 * @param initialX
	 *            the initial x position of the cursor.
	 * @param initialY
	 *            the initial y position of the cursor.
	 */
	public Cursor(/*unitfor(R)*/Rectangle rectangle, int initialX, int initialY) {
		this.rectangle = new Rectangle/*R=this.C*/(rectangle);
		this.x = initialX;
		this.y = initialY;
	}
	
//	protected Cursor(/*unitfor(C)*/Cursor c){
//		this.rectangle = c.rectangle;
//		this.x = c.getX();
//		this.y = c.getY();
//	}

	//not needed with aj ; added above constructor to allow reconstruction with proper alias
//	protected void setAW(AbstractWindow w) {
//		if (aw != null)
//			throw new RuntimeException(
//					"Cursor should be associated with no more than one window.");
//		aw = w;
//		this.rectangle.setAW(w);
//	}

	/**
	 * Retrieves the current X coordinate of the cursor.
	 * 
	 * @return the current X coordinate of the cursor.
	 */
	public int getX() {
		return x;
	}

	/**
	 * Retrieves the current Y coordinate of the cursor.
	 * 
	 * @return the current Y coordinate of the cursor.
	 */
	public int getY() {
		return y;
	}

	/**
	 * Changes the current X coordinate of the cursor.
	 * 
	 * @param x
	 *            the new X coordinate of the cursor.
	 * 
	 * @exception InconsistentCoordinateException
	 *                if <code>x</code> is not consistent.
	 */
	public void setX(int x) {
		setXY(x, this.y);
	}

	/**
	 * Changes the current Y coordinate of the cursor.
	 * 
	 * @exception InconsistentCoordinateException
	 *                if <code>y</code> is not consistent.
	 * @param y
	 *            the new Y coordinate of the cursor.
	 */
	public void setY(int y) {
		setXY(this.x, y);
	}

	/**
	 * Changes both the current X and current Y coordinates of the cursor.
	 * 
	 * @param x
	 *            the new X coordinate of the cursor.
	 * @param y
	 *            the new Y coordinate of the cursor.
	 * 
	 * @exception InconsistentCoordinateException
	 *                if <code>x</code> or <code>y</code> is not consistent.
	 */
	public void setXY(int x, int y) {
//aj remove
//		if (aw == null)
//			throw new RuntimeException(
//					"Coordinates cannot be set on a cursor not associated with a window.");

//		synchronized (aw) {
			int width = rectangle.getWidth();
			if (x < 0 || x >= width) {
				throw new InconsistentCoordinateException("x: " + x
						+ " is not " + "consistent with width:" + width);
			}
			int height = rectangle.getHeight();
			if (y < 0 || y >= height) {
				throw new InconsistentCoordinateException("y: " + y
						+ " is not " + "consistent with height:" + height);
			}
			this.x = x;
			this.y = y;
//		}
	}

	/**
	 * Changes cursor visibility.
	 * 
	 * @param visible
	 *            the cursor visibility.
	 * 
	 * @exception IllegalArgumentException
	 *                if <code>visible</code> is not consistent.
	 */
	public void setCursorVisibility(int visible) {
		if (visible < INVISIBLE || visible > VERY_VISIBLE) {
			throw new IllegalArgumentException("visible: " + visible
					+ "should be between " + INVISIBLE + " and " + VERY_VISIBLE);
		}
		this.visibility = visible;
//		Screen.setCursorVisibilityOnScreenIfInited(visible);
//		Screen.getScreen().setCursorVisibilityOnScreen(visible);
	}

	/**
	 * Retrieves the cursor visibility.
	 * 
	 * @return the cursor visibility.
	 */
	public int getCursorVisibility() {
		return visibility;
	}

	/** Cursor is not visible. */
	public static final int INVISIBLE = 0;

	/** Cursor is visible. */
	public static final int NORMAL = 1;

	/** Cursor is very visible. */
	public static final int VERY_VISIBLE = 2;
}
