/*  Jcurzez - A free Java implementation of curses
 *  Copyright (C) 2001, 2006  Thomas G. Girard <thomas.g.girard@free.fr>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.nongnu.savannah.jcurzez;

/**
 * A rectangle is defined by its bounds ({@link #top}, {@link #left},
 * {@link #right}, {@link #bottom}) a {@link #fillingPen background pen}, and a
 * {@link #drawingPen drawing pen}. The first pen will be used to fill in the
 * rectangle when new room is created. This is a kind of background pen. The
 * second one is the one you should change when you want to write a new thing on
 * the window, using different colors and/or attributes.
 * <p>
 * This class doesn't provide drawing primitives, since drawing is delegated to
 * {@link jcurzez.AbstractWindow AbstractWindow}.
 * 
 * @author Thomas Girard
 * @version 1.0.0
 * 
 * @see jcurzez.BackgroundColor
 */
public class Rectangle {
	// Those fields are not final because maybe a rectangle could be resized or
	// moved.

	/*atomicset(R)*/
	
	/** top coordinate of this rectangle. */
	/*atomic(R)*/protected int top;

	/** left coordinate of this rectangle. */
	/*atomic(R)*/protected int left;

	/** right coordinate of this rectangle. */
	/*atomic(R)*/protected int right;

	/** bottom coordinate of this rectangle. */
	/*atomic(R)*/protected int bottom;

	// aj removing final for ease of construction workaround
	/** pen used to fill in this rectangle. */
	protected /*final*/ FillingPen fillingPen/*P=this.R*/;

	/** pen used to draw new cells in this rectangle. */
	protected /*final*/ Pen drawingPen/*P=this.R*/;

	// aj remove
	//AbstractWindow aw;
	
	// aj add to allow reconstruction with proper aliasing
	protected Rectangle(/*unitfor(R)*/Rectangle r){
		initCoords(r.getLeft(),r.getTop(),r.getRight(),r.getBottom());
		initPens(r.fillingPen,r.drawingPen);
		//this (r.left,r.top,r.right,r.bottom,r.fillingPen,r.drawingPen);
	}

	/**
	 * Creates a new Rectangle with the specified bounds, and a default pen for
	 * drawing and filling.
	 * 
	 * @param left
	 *            left coordinate.
	 * @param top
	 *            top coordinate.
	 * @param right
	 *            right coordinate.
	 * @param bottom
	 *            bottom coordinate.
	 */
	public Rectangle(int left, int top, int right, int bottom) {
		initCoords(left,top,right,bottom);
		initPens(new Pen/*P=this.R*/(), new Pen/*P=this.R*/());
		//this(left, top, right, bottom, new Pen/*P=this.R*/(), new Pen/*P=this.R*/());
	}

	/**
	 * Creates a new Rectangle with the specified bounds, specified pen fo
	 * filling, and a default pen for drawing.
	 * 
	 * @param left
	 *            left coordinate.
	 * @param top
	 *            top coordinate.
	 * @param right
	 *            right coordinate.
	 * @param bottom
	 *            bottom coordinate.
	 * @param fillingPen
	 *            the pen used to fill cells.
	 */
	public Rectangle(int left, int top, int right, int bottom, /*unitfor(P)*/Pen fillingPen) {
		initCoords(left,top,right,bottom);
		initPens(fillingPen, new Pen/*P=this.R*/());
		//this(left, top, right, bottom, fillingPen, new Pen/*P=this.R*/());
	}

	/**
	 * Creates a new Rectangle with the specified bounds, and specified
	 * background color.
	 * 
	 * @param left
	 *            left coordinate.
	 * @param top
	 *            top coordinate.
	 * @param right
	 *            right coordinate.
	 * @param bottom
	 *            bottom coordinate.
	 * @param fillingPen
	 *            the pen to fill the rectangle.
	 * @param drawingPen
	 *            the pen to draw new cells.
	 */
	public Rectangle(int left, int top, int right, int bottom, /*unitfor(P)*/ Pen fillingPen,
			/*unitfor(P)*/ Pen drawingPen) {
		initCoords(left,top,right,bottom);
		initPens(fillingPen,drawingPen);
	}

	private void initCoords(int left, int top, int right, int bottom){
		this.top = top;
		this.left = left;
		this.bottom = bottom;
		this.right = right;
	}
	
	public void incrementCoords(int left, int top, int right, int bottom){
		this.top += top;
		this.left += left;
		this.bottom += bottom;
		this.right += right;
	}
	
	private void initPens( /*unitfor(P)*/Pen fillingPen, /*unitfor(P)*/Pen drawingPen){
		this.fillingPen = new FillingPen/*P=this.R*/(fillingPen);
		this.drawingPen = new Pen/*P=this.R*/(drawingPen);
	}
	
	// aj remove
//	void setAW(AbstractWindow w) {
//		if (aw != null)
//			throw new RuntimeException(
//					"Rectangle should not be associated with more than one window.");
//		aw = w;
//		fillingPen.aw = w;
//		drawingPen.aw = w;
//	}

	/**
	 * Returns the left coordinate.
	 * 
	 * @return the left coordinate.
	 */
	public int getLeft() {
		return left;
	}

	/**
	 * Returns the top coordinate.
	 * 
	 * @return the top coordinate.
	 */
	public int getTop() {
		return top;
	}

	/**
	 * Returns the right coordinate.
	 * 
	 * @return the right coordinate.
	 */
	public int getRight() {
		return right;
	}

	/**
	 * Returns the bottom coordinate.
	 * 
	 * @return the bottom coordinate.
	 */
	public int getBottom() {
		return bottom;
	}

	/**
	 * Returns the height of the rectangle.
	 * 
	 * @return the height of the rectangle.
	 */
	public int getHeight() {
//		synchronized (aw) {
			return bottom - top;
//		}
	}

	/**
	 * Returns the width of the rectangle.
	 * 
	 * @return the width of the rectangle.
	 */
	public int getWidth() {
//		synchronized (aw) {
			return right - left;
//		}
	}

	/**
	 * Returns the drawing pen of this rectangle.
	 * 
	 * @return {@link #drawingPen}.
	 */
	public Pen getDrawingPen() {
		return drawingPen;
	}

	/**
	 * Returns the filling pen.
	 * 
	 * @return {@link #fillingPen}.
	 */
	public Pen getFillingPen() {
		return fillingPen;
	}

	/**
	 * Returns a representation of <code>this</code>.
	 * 
	 * @return a representation of <code>this</code>.
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer("[left=");
		sb.append(left).append(",top=").append(top).append(",right=");
		sb.append(right).append(",bottom=").append(bottom).append(']');
		return sb.toString();
	}
}
