/*  Jcurzez - A free Java implementation of curses
 *  Copyright (C) 2001-2007  Thomas Girard <thomas.g.girard@free.fr>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.nongnu.savannah.jcurzez;

/**
 * <code>AbstractWindow</code> is the base class of {@link Area} and
 * {@link Screen}. <br>
 * An <code>AbstractWindow</code> is defined by:
 * <ul>
 * <li>a {@link jcurzez.Rectangle rectangle}: it contains the bounds of this
 * window</li>
 * <li>a {@link jcurzez.Cursor cursor}: it contains the current position of the
 * cursor in the window</li>
 * <li>two {@link jcurzez.Pen pens}: one for drawing characters that are
 * {@link #printChar printed}, the other one for printing cells created when
 * {@link #deleteChar deleting}, {@link #insertChar inserting} characters and
 * lines.</li>
 * <li>a {@link #id unique id}</li>
 * <li>two special boolean properties:
 * <ul>
 * <li>{@link #linewrap}, which allows a line too long to continue on the next
 * line</li>
 * <li>{@link #scroll}, which allows the window to scroll when text reaches the
 * bottom</li>
 * </ul>
 * </li>
 * </ul>
 * <p>
 * Every <code>AbstractWindow</code> has a {@link #buffer} that is used for all
 * operations. This buffer can be shared with the one from its parent window
 * (see {@link #AbstractWindow(AbstractWindow,Cursor,boolean,boolean,boolean)}).
 * <p>
 * One has to invoke {@link #refresh} to sync the content of the window with the
 * buffer.
 * <p>
 * <code>AbstractWindow</code> is the biggest piece of <b>Jcurzez</b>. Bugs for
 * sure are here.
 * 
 * @author Thomas Girard
 * @version 1.0.0
 * 
 * @see jcurzez.Rectangle
 * @see jcurzez.Cursor
 * @see jcurzez.Pen
 * @see jcurzez.Screen
 * @see jcurzez.Area
 */
public abstract class AbstractWindow {
	/** Counter for next available window number. */
	private static long nextId = 0L;

	/** Tabulation size: defaults to 8. */
	private static final int tabSize = Integer.getInteger(
			"jcurzez.term.tabsize", 8).intValue();

	/** Number of this window. */
	protected final long id;

	/** Parent window. */
	protected final AbstractWindow parent;

	/** Cursor attached to this window (the one for drawing new characters). */
	protected final Cursor cursor;

	/**
	 * Indicates if lines are wrapped. This can not be altered after creation.
	 */
	protected final boolean linewrap;

	/** Would the window automatically scroll if needed ? */
	protected final boolean scroll;

	/**
	 * The buffer we use to perform all operations. Use {@link #refresh} to sync
	 * the content of the window with this buffer.
	 */
	private final Cell[][] buffer;

	/** The lock on the buffer. */
	private final Object bufferLock;

	/** <code>true</code> if {@link #buffer} is shared with {@link #parent}. */
	private final boolean isBufferShared;

	/**
	 * Setting this flag will cause printing strings to the buffer to go very
	 * slowly so we can see what is allowed to run concurrently.
	 */
	public boolean stressConcurrency = false;

	/**
	 * Constructs a new window with specified parent, rectangle and cursor. This
	 * window will not share the buffer of its parent.
	 * 
	 * @param parent
	 *            the parent window.
	 * @param cursor
	 *            the cursor attached to this window.
	 * @param linewrap
	 *            if lines will be wrapped.
	 * @param scroll
	 *            if the window should scroll.
	 * 
	 * @exception InconsistentCoordinateException
	 *                if the window being created is not contained in its
	 *                parent.
	 */
	protected AbstractWindow(AbstractWindow parent, Cursor cursor,
			boolean linewrap, boolean scroll) {
		this(parent, cursor, linewrap, scroll, false);
	}

	/**
	 * Constructs a new window with specified parent, rectangle and cursor.
	 * 
	 * @param parent
	 *            the parent window.
	 * @param cursor
	 *            the cursor attached to this window.
	 * @param linewrap
	 *            if lines will be wrapped.
	 * @param scroll
	 *            if the window should scroll.
	 * @param share
	 *            if the new window should share its buffer with its parent.
	 * 
	 * @exception InconsistentCoordinateException
	 *                if the window being created is not contained in its
	 *                parent.
	 * @exception IllegalArgumentException
	 *                if <code>share</code> is <code>true</code> and
	 *                <code>parent</code> is <code>null</code>.
	 */
	protected AbstractWindow(AbstractWindow parent, Cursor cursor,
			boolean linewrap, boolean scroll, boolean share) {
		this.parent = parent;
		this.cursor = cursor;
		cursor.setAW(this);
		this.linewrap = linewrap;
		this.scroll = scroll;
		this.id = nextId++;

		final int width = cursor.rectangle.getWidth();
		final int height = cursor.rectangle.getHeight();

		// We check that this new window is within its parent, if it exists.
		if (parent != null) {
			if ((cursor.rectangle.left < parent.cursor.rectangle.left)
					|| (cursor.rectangle.top < parent.cursor.rectangle.top)
					|| (cursor.rectangle.right > parent.cursor.rectangle.right)
					|| (cursor.rectangle.bottom > parent.cursor.rectangle.bottom)) {
				throw new InconsistentCoordinateException("The window being "
						+ "created is not " + "contained in its parent");
			}
		} else if (share) {
			throw new IllegalArgumentException("Parent cannot be null when "
					+ "buffer sharing is required");
		}

		this.isBufferShared = share;

		if (share) {
			bufferLock = parent.bufferLock;
		} else {
			bufferLock = new Object();
		}

		Object dontDeadlockLock = parent==null?bufferLock:parent;
		synchronized (dontDeadlockLock) {
			synchronized (bufferLock) {
				if (share) {
					buffer = parent.getCellsAt(cursor.rectangle.left
							- parent.cursor.rectangle.left,
							cursor.rectangle.top - parent.cursor.rectangle.top,
							width, height);
				} else {
					// Initialize the buffer.
					buffer = new Cell[height][width];

					Cell blank = cursor.rectangle.fillingPen.blankCell;

					for (int i = 0; i < height; i++) {
						for (int j = 0; j < width; j++) {
							buffer[i][j] = new Cell(blank);
						}
					}
				}
			}
		}
	}

	/**
	 * Returns the parent window of <code>this</code>. <code>null</code> is the
	 * parent of the {@link jcurzez.Screen screen}.
	 * 
	 * @return the parent window of <code>this</code>.
	 */
	public AbstractWindow getParent() {
		return parent;
	}

	/**
	 * Returns the rectangle contained within <code>this</code>.
	 * 
	 * @return the rectangle contained within <code>this</code>.
	 */
	public Rectangle getRectangle() {
		return cursor.rectangle;
	}

	/**
	 * Returns the cursor attached to <code>this</code>.
	 * 
	 * @return the cursor of <code>this</code>.
	 */
	public Cursor getCursor() {
		return cursor;
	}

	/**
	 * Returns the size of the tabulation.
	 * 
	 * @return {@link #tabSize}.
	 */
	public static int getTabulationSize() {
		return tabSize;
	}

	/**
	 * Returns the id of this window.
	 * 
	 * @return {@link #id}.
	 */
	public long getId() {
		return id;
	}

	/**
	 * Returns <code>true</code> if lines are wrapped.
	 * 
	 * @return <code>true</code> if lines are wrapped.
	 */
	public boolean hasLinesWrapped() {
		return linewrap;
	}

	/**
	 * Returns <code>true</code> if the window would scroll if needed.
	 * 
	 * @return <code>true</code> if the window has the automatic scrolling
	 *         feature.
	 */
	public boolean wouldScroll() {
		return scroll;
	}

	/**
	 * Inserts a character at current cursor position, with the filling pen.
	 * 
	 * @param c
	 *            the character to insert.
	 */
	public synchronized void insertChar(int c) {
		synchronized (bufferLock) {
			_insertChar(c);
		}
	}

	/**
	 * Inserts a character at current cursor position, with the filling pen.
	 * <i>This is the unsynchronized version</i>.
	 * 
	 * @param c
	 *            the character to insert.
	 */
	private void _insertChar(int c) {
		int width = cursor.rectangle.getWidth();
		char character = (char) c;

		switch (character) {
		case '\f':
			_clear();
			break;
		case '\b':
			backspace();
			break;
		case '\t':
			for (int i = 0; i < tabSize; i++) {
				_insertChar((int) ' ');
			}
			break;
		case '\r':
			cursor.x = 0;
			break;
		case '\n':
			int len = width - cursor.x;
			Cell[][] restOfTheLine = copyCellsAt(cursor.x, cursor.y, len, 1);
			_clearToEndOfLine();

			cursor.x = 0;
			if (increaseY())
				_insertLine();

			for (int i = 0; i < len; i++) {
				restOfTheLine[0][i].copyInto(buffer[cursor.y][cursor.x + i]);
			}

			cursor.x += len;
			break;
		default:
			for (int i = width - 2; i >= cursor.x; i--) {
				buffer[cursor.y][i].copyInto(buffer[cursor.y][i + 1]);
			}
			buffer[cursor.y][cursor.x].copyFrom(c,
					cursor.rectangle.fillingPen.attribute,
					cursor.rectangle.fillingPen.background,
					cursor.rectangle.fillingPen.foreground);
		}
	}

	/**
	 * What to do when a <code>'\b'</code> is {@link #printChar}'ed or
	 * {@link #insertChar}'ed.
	 */
	private void backspace() {
		cursor.x--;
		if (cursor.x < 0) {
			if (cursor.y - 1 < 0) {
				cursor.x++;
				// can't do anything, skip this backspace.
				return;
			}
			cursor.y--;
			cursor.x = cursor.rectangle.getWidth() - 1;
		}
		// already has the lock.
		_deleteChar();
	}

	/**
	 * Deletes char at current position.
	 */
	public synchronized void deleteChar() {
		synchronized (bufferLock) {
			_deleteChar();
		}
	}

	/**
	 * Deletes char at current position. <i>This is the unsynchronized
	 * version</i>.
	 */
	private void _deleteChar() {
		int width = cursor.rectangle.getWidth();

		for (int i = cursor.x; i < width - 1; i++) {
			buffer[cursor.y][i + 1].copyInto(buffer[cursor.y][i]);
		}

		cursor.rectangle.fillingPen.blankCell
				.copyInto(buffer[cursor.y][width - 1]);
	}

	/**
	 * Inserts a blank line at current position, with default pen.
	 */
	public synchronized void insertLine() {
		synchronized (bufferLock) {
			_insertLine();
		}
	}

	/**
	 * Inserts a blank line at current position, with default pen. <i>This is
	 * the unsynchronized version</i>.
	 */
	private void _insertLine() {
		int height = cursor.rectangle.getHeight();
		int width = cursor.rectangle.getWidth();

		for (int i = height - 2; i >= cursor.y; i--) {
			for (int j = 0; j < width; j++) {
				buffer[i][j].copyInto(buffer[i + 1][j]);
			}
		}

		for (int i = 0; i < width; i++) {
			cursor.rectangle.fillingPen.blankCell.copyInto(buffer[cursor.y][i]);
		}
	}

	/**
	 * Deletes the current line.
	 */
	public synchronized void deleteLine() {
		synchronized (bufferLock) {
			_deleteLine();
		}
	}

	/**
	 * Deletes the current line. <i>This is the unsynchronized version</i>.
	 */
	private void _deleteLine() {
		int height = cursor.rectangle.getHeight();
		int width = cursor.rectangle.getWidth();

		for (int i = cursor.y; i < height - 1; i++) {
			for (int j = 0; j < width; j++) {
				buffer[i + 1][j].copyInto(buffer[i][j]);
			}
		}

		for (int i = 0; i < width; i++) {
			cursor.rectangle.fillingPen.blankCell
					.copyInto(buffer[height - 1][i]);
		}
	}

	/**
	 * Clears the current window.
	 */
	public synchronized void clear() {
		synchronized (bufferLock) {
			_clear();
		}
	}

	/**
	 * Clears the current window. <i>This is the non synchronized version</i>.
	 */
	private void _clear() {
		Cell blank = cursor.rectangle.fillingPen.blankCell;

		cursor.x = 0;
		cursor.y = 0;
		for (int i = 0, height = cursor.rectangle.getHeight(), width = cursor.rectangle
				.getWidth(); i < height; i++) {
			for (int j = 0; j < width; j++) {
				blank.copyInto(buffer[i][j]);
			}
		}
	}

	/**
	 * Emits a beep. This method will be provided by an implementation of
	 * <code>ScreenPeer</code>.
	 * 
	 * @see jcurzez.peer.ScreenPeer#beep
	 */
	public abstract void beep();

	/**
	 * Prints a cell on the <b>physical screen</b> at absolute position
	 * <code>(x, y)</code>. This method will be provided by an implementation of
	 * <code>ScreenPeer</code>.
	 * 
	 * @param x
	 *            the x coordinate where to print the cell.
	 * @param y
	 *            the y coordinate where to print the cell.
	 * @param cell
	 *            the cell to print.
	 * 
	 * @see jcurzez.peer.ScreenPeer#printCell
	 */
	protected abstract void printCell(int x, int y, final Cell cell);

	/**
	 * Moves the cursor on the <b>physical screen</b> to position
	 * <code>(x, y)</code>. This method will be provided by an implementation of
	 * <code>ScreenPeer</code>.
	 * 
	 * @param x
	 *            the x coordinate where to move the cursor.
	 * @param y
	 *            the y coordinate where to move the cursor.
	 * 
	 * @see jcurzez.peer.ScreenPeer#gotoXY
	 */
	protected abstract void gotoXY(int x, int y);

	/**
	 * Updates the <b>physical screen</b>. This method will be provided by an
	 * implementation of <code>ScreenPeer</code>.
	 * 
	 * @see jcurzez.peer.ScreenPeer#update
	 */
	protected abstract void update();

	/**
	 * Clears the content of the window, from current position to the end of the
	 * screen, using the filling pen for filling new cells.
	 */
	public synchronized void clearToBottom() {
		Cell blank = cursor.rectangle.fillingPen.blankCell;

		synchronized (bufferLock) {
			int j = cursor.x;
			for (int i = cursor.y, height = cursor.rectangle.getHeight(), width = cursor.rectangle
					.getWidth(); i < height; i++) {
				while (j < width) {
					blank.copyInto(buffer[i][j]);
					j++;
				}
				j = 0;
			}
		}
	}

	/**
	 * Clears the content of the line, starting from current position, using the
	 * filling pen for new cells.
	 */
	public synchronized void clearToEndOfLine() {
		synchronized (bufferLock) {
			_clearToEndOfLine();
		}
	}

	/**
	 * Clears the content of the line, starting from current position, using the
	 * filling pen for new cells. <i>This is the non synchronized version</i>.
	 */
	private void _clearToEndOfLine() {
		Cell blank = cursor.rectangle.fillingPen.blankCell;

		for (int i = cursor.x, width = cursor.rectangle.getWidth(); i < width; i++) {
			blank.copyInto(buffer[cursor.y][i]);
		}
	}

	/**
	 * Inserts specified text <code>text</code> at current position, using the
	 * filling pen.
	 * 
	 * @param text
	 *            the text to insert.
	 */
	public synchronized void insertString(String text) {
		synchronized (bufferLock) {
			for (int i = text.length() - 1; i >= 0; i--)
				insertChar((int) text.charAt(i));
		}
	}

	/**
	 * Puts <code>c</code> with given decoration at current position in
	 * {@link #buffer}.
	 * 
	 * @param c
	 *            the nwe character for the cell.
	 * @param fg
	 *            the new foreground color for the cell.
	 * @param bg
	 *            the new background color for the cell.
	 * @param attr
	 *            the new attrivute for the cell.
	 */
	private final void internalPutChar(int c, ForegroundColor fg,
			BackgroundColor bg, Attribute attr) {
		buffer[cursor.y][cursor.x].copyFrom(c, attr, bg, fg);
	}

	/**
	 * Prints specified character <code>c</code> at current position, using the
	 * drawing pen.
	 * 
	 * @param c
	 *            the character to print.
	 */
	public synchronized void printChar(int c) {
		synchronized (bufferLock) {
			_printChar(c);
		}
	}

	/**
	 * Prints specified character <code>c</code> at current position, using the
	 * drawing pen. <i>This is the unsynchronized version</i>.
	 * 
	 * @param c
	 *            the character to print.
	 */
	private void _printChar(int c) {
		char character = (char) c;

		final ForegroundColor fg = cursor.rectangle.drawingPen
				.getForegroundColor();
		final BackgroundColor bg = cursor.rectangle.drawingPen
				.getBackgroundColor();
		final Attribute at = cursor.rectangle.drawingPen.getAttribute();

		switch (character) {
		case '\f':
			_clear();
			break;
		case '\b':
			backspace();
			break;
		case '\t':
			for (int i = 0; i < tabSize; i++) {
				internalPutChar((int) ' ', fg, bg, at);
				increaseX();
			}
			break;
		case '\r':
			cursor.x = 0;
			break;
		case '\n':
			cursor.x = 0;
			increaseY();
			break;
		default:
			internalPutChar(c, fg, bg, at);
			increaseX();
		}

		if (stressConcurrency)
			try {
				Thread.sleep(10);
			} catch (InterruptedException e) {
			}
	}

	/**
	 * Prints specified String <code>text</code> at current position, using the
	 * drawing pen.
	 * 
	 * @param text
	 *            the text to print.
	 */
	public synchronized void printString(String text) {
		if (!linewrap) {
			text = removeExtraChars(text);
		}

		synchronized (bufferLock) {
			for (int i = 0; i < text.length(); i++) {
				_printChar((int) text.charAt(i));
			}
		}
	}

	/**
	 * Redraws the window.
	 */
	public synchronized void refresh() {
		if (isBufferShared) {
			// if the buffer is shared, invoking parent's
			// refresh will do the job.
			parent.refresh();
			gotoXY(cursor.rectangle.left + cursor.x, cursor.rectangle.top
					+ cursor.y);
		} else {
			// /cat.debug("Requesting refresh");
			Cell cell = null;
			for (int i = 0, w = cursor.rectangle.getWidth(), h = cursor.rectangle
					.getHeight(); i < h; i++) {
				for (int j = 0; j < w; j++) {
					synchronized (bufferLock) {
						cell = new Cell(buffer[i][j]);
						buffer[i][j].hasChanged = false;
					}

					// doing refresh one cell at a time, releasing bufferlock,
					// may allow user to observe
					// operations out of the order in which a thread performs
					// them: T1:refreshPrintOldCell1, T2:updateCell1,
					// T2:updateCell2, T1:refreshPrintNewCell2
					// (Must have shared buffer and refresh called on different
					// AbsWindow from update)

					if (cell.hasChanged) {
						printCell(cursor.rectangle.left + j,
								cursor.rectangle.top + i, cell);
						
						// synchronized(bufferLock) {
						// buffer[i][j].hasChanged = false;
						// }
					}
				}
			}
			// /cat.debug("Refresh done");

			// A way for the peer to perform optimization.
			update();

			// return to where we were before invoking refresh.
			gotoXY(cursor.rectangle.left + cursor.x, cursor.rectangle.top
					+ cursor.y);
		}
	}

	/**
	 * Remove extra chars from <code>text</code>. When {@link #linewrap} is set
	 * to <code>false</code>, some chars won't appear, so they are removed.
	 * 
	 * @param text
	 *            the text to modify.
	 * @return modified <code>text</code>.
	 */
	protected String removeExtraChars(String text) {
		int len = text.length();
		int maxPrintableChars = Math.min(len, cursor.rectangle.getWidth()
				- cursor.x);

		StringBuffer sb = new StringBuffer();
		int nbPrintableCharsCopied = 0, index = 0;

		while (nbPrintableCharsCopied < maxPrintableChars && index < len) {
			char c = text.charAt(index++);
			sb.append(c);
			if (!Character.isISOControl(c)) {
				nbPrintableCharsCopied++;
			}
		}
		return sb.toString();
	}

	/**
	 * Moves the cursor one character down. If needed, and {@link #scroll} is
	 * <code>true</code>, then we insert a new line. Otherwise, the cursor will
	 * remain at the same position, and we return <code>false</code>.
	 * 
	 * @return <code>true</code> if the y position could be changed.
	 */
	private boolean increaseY() {
		if (cursor.y + 1 == cursor.rectangle.getHeight()) {
			if (scroll) {
				int oldY = cursor.y;
				cursor.y = 0;
				_deleteLine();
				cursor.y = oldY;
				return true;
			}
			return false;
		} else {
			cursor.y++;
		}
		return true;
	}

	/**
	 * Moves the cursor one character right. Updates the y coordinate if x reach
	 * the right bound. If this is not possible, then the cursor is left where
	 * it is.
	 */
	private void increaseX() {
		cursor.x++;
		if (cursor.x == cursor.rectangle.getWidth()) {
			int oldX = cursor.x - 1;
			cursor.x = 0;
			if (!increaseY()) {
				cursor.x = oldX;
			}
		}
	}

	/**
	 * Tests equality of <code>this</code> with <code>o</code>. Two
	 * <code>AbstractWindow</code> are said to be equal if they have the same
	 * {@link #id}.
	 * 
	 * @param o
	 *            the object we want to test for equality with <code>this</code>
	 *            .
	 * @return <code>true</code> if <coe>this</code> and <code>o</code> have the
	 *         same <code>id</code>.
	 */
	public boolean equals(Object o) {
		if (o == this)
			return true;
		if (o instanceof AbstractWindow) {
			return id == ((AbstractWindow) o).id;
		}
		return false;
	}

	/**
	 * Returns the hash code of <code>this</code>. It is computed using its
	 * {@link #id}.
	 * 
	 * @return the hash code to use for <code>this</code>.
	 */
	public int hashCode() {
		return (int) (id ^ (id >>> 32));
	}

	/**
	 * Returns a string representation of <code>this</code>. It has the form [
	 * x,y p1 p2 ] where p1 is the string representation of the drawing pen and
	 * p2 the string representation of the filling pen.
	 * 
	 * @return the string representation of <code>this</code>.
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer("[ ");
		sb.append(cursor.x).append(',').append(cursor.y);
		sb.append(' ').append(cursor.rectangle.drawingPen);
		sb.append(' ').append(cursor.rectangle.fillingPen);
		return sb.append(" ]").toString();
	}

	/**
	 * Marks all characters as changed so that the next call to {@link #refresh}
	 * will have the whole window redisplayed.
	 */
	public synchronized void touch() {
		synchronized (bufferLock) {
			for (int i = 0, w = cursor.rectangle.getWidth(), h = cursor.rectangle
					.getHeight(); i < h; i++) {
				for (int j = 0; j < w; j++) {
					buffer[i][j].hasChanged = true;
				}
			}
		}
	}

	/**
	 * Checks coordinates for both {@link #copyCellsAt} and {@link #getCellsAt}.
	 */
	private void checkAllCoordinates(int x, int y, int width, int height) {
		final int w = cursor.rectangle.getWidth();
		final int h = cursor.rectangle.getHeight();

		if (x < 0 || x >= w) {
			throw new InconsistentCoordinateException("x: " + x
					+ " is not consistent with " + "width: " + w);
		}
		if (y < 0 || y >= h) {
			throw new InconsistentCoordinateException("y: " + y
					+ " is not consistent with " + "height: " + h);
		}

		if (width < 0) {
			throw new IllegalArgumentException("width: " + width
					+ " must be >= 0");
		}

		if (height < 0) {
			throw new IllegalArgumentException("height: " + height
					+ " must be >= 0");
		}

		if (x + width > w) {
			throw new ArrayIndexOutOfBoundsException("width: " + width
					+ " is greater than possible: " + (w - x));
		}
		if (y + height > h) {
			throw new ArrayIndexOutOfBoundsException("height: " + height
					+ " is greater than possible: " + (h - y));
		}
	}

	/**
	 * Retrieves the rectangle of cells starting at (<code>x</code>,
	 * <code>y</code>). The retrieved array has a width of <code>width</code>
	 * and a height of <code>height</code>. The cells are <b>copied</b>. The
	 * first subscript of the returned array is the y coordinate.
	 * 
	 * @param x
	 *            the x coordinate from where to retrieve cells.
	 * @param y
	 *            the y coordinate from where to retrieve cells.
	 * @param width
	 *            the width of the array to get.
	 * @param height
	 *            the height of the array to get.
	 * @exception InconsistentCoordinateException
	 *                if either <code>x</code> or <code>y</code> is not
	 *                consistent.
	 * @exception ArrayIndexOutOfBoundsException
	 *                if <code>width</code> or <code>height</code> is not
	 *                consistent.
	 */
	protected Cell[][] copyCellsAt(int x, int y, int width, int height) {
		final int w = cursor.rectangle.getWidth();
		final int h = cursor.rectangle.getHeight();

		checkAllCoordinates(x, y, width, height);

		synchronized (bufferLock) {
			Cell[][] cells = new Cell[height][width];
			for (int i = 0; i < height; i++) {
				for (int j = 0; j < width; j++) {
					cells[i][j] = new Cell(buffer[y + i][x + j]);
				}
			}
			return cells;
		}
	}

	/**
	 * Retrieves the rectangle of cells starting at (<code>x</code>,
	 * <code>y</code>). The retrieved array has a width of <code>width</code>
	 * and a height of <code>height</code>. The cells are <b>referenced</b>. The
	 * first subscript of the returned array is the y coordinate.
	 * 
	 * @param x
	 *            the x coordinate from where to retrieve cells.
	 * @param y
	 *            the y coordinate from where to retrieve cells.
	 * @param width
	 *            the width of the array to get.
	 * @param height
	 *            the height of the array to get.
	 * @exception InconsistentCoordinateException
	 *                if either <code>x</code> or <code>y</code> is not
	 *                consistent.
	 * @exception ArrayIndexOutOfBoundsException
	 *                if <code>width</code> or <code>height</code> is not
	 *                consistent.
	 */
	protected Cell[][] getCellsAt(int x, int y, int width, int height) {
		final int w = cursor.rectangle.getWidth();
		final int h = cursor.rectangle.getHeight();

		checkAllCoordinates(x, y, width, height);

		synchronized (bufferLock) {
			Cell[][] cells = new Cell[height][width];
			for (int i = 0; i < height; i++) {
				for (int j = 0; j < width; j++) {
					cells[i][j] = buffer[y + i][x + j];
				}
			}
			return cells;
		}
	}
}
